/*
 * Copyright (C) 2006 Universitat d'Alacant / Universidad de Alicante
 * author: Felipe Snchez-Martnez
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <apertium/LexTor.H>
#include <apertium/StringUtils.H>

#include <algorithm>
#include <string>
#include <cmath>


bool LexTor::debug;

LexTor::LexTor() {
  lextor_data=NULL;
}
  
LexTor::LexTor(const LexTor& lt) {
  lextor_data=lt.lextor_data;
}
  
LexTor::~LexTor() {
}

void 
LexTor::set_lextor_data(LexTorData* ltd) {
  lextor_data=ltd;
}

void 
LexTor::trainwrd(istream& is, int left, int right, double weigth_exponent) {
  if (lextor_data==NULL) {
    cerr<<"Error in LexTor::trainwrd, you must call set_lextor_data before training\n";
    exit(EXIT_FAILURE);
  }

  lextor_data->ensure_stopwords_ok();

  cerr<<"Number of words to take into account on the left side: "<<left<<"\n";
  cerr<<"Number of words to take into account on the right side: "<<right<<"\n";

  set<string> words2workwith=lextor_data->get_words();
  set<string>::iterator itword;

  cerr<<"Words to work with:\n";
  for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
    cerr<<*itword<<"\n";
  }
  cerr<<"\n";

  is.clear();
  is.seekg(ios::beg);

  int nw=0;

  map<string, map<string, COUNT_DATA_TYPE> > context;
  deque<string> buffer;
  unsigned word_index= (unsigned)left;

  unsigned buffer_max_size=(unsigned)(left+1+right);
  
  LexTorWord *ltword;
  ltword=LexTorWord::next_word(is);
  while(ltword!=NULL) {
    if ((++nw%250000)==0)
      cerr<<nw<<" words processed\n";
      
    //cerr<<"Word read from corpus: "<<ltword->get_word_string()
    //	<<", reduced: "<<lextor_data->reduce(ltword->get_word_string())<<"\n";
    //getchar();

    string reduced_word=lextor_data->reduce(ltword->get_word_string());

    if (!lextor_data->is_stopword(reduced_word)) {
      if (buffer.size()>=buffer_max_size) {
	buffer.pop_front();
      }
      buffer.push_back(reduced_word);

      //The buffer is already full
      if (buffer.size()==buffer_max_size) {
	for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
	  if (buffer[word_index]==(*itword)) {
	    int distance=(-1)*left;
	    for(unsigned i=0; i<buffer.size(); i++) {
	      if ((i!=word_index) && (buffer[i]!=(*itword))) {
		context[*itword][buffer[i]]+=1.0/pow(fabs((double)distance),weigth_exponent);
	      }
	      distance++;
	    }
	    break;
	  }
	}
      }
    }

    delete ltword;
    ltword=LexTorWord::next_word(is);
  }

  cerr<<"Corpus has "<<nw<<" words\n";

  //All co-occurrences have been collected. We need to filter them
  //so as to take into account only the n most frequents
  for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
    PairStringCountComparer comparer;
    vector<pair<string, COUNT_DATA_TYPE> > context_v;
    map<string, COUNT_DATA_TYPE>::iterator itm;

    while(context[*itword].size()>0) {
      itm=context[*itword].begin();
      context_v.push_back(*itm);
      context[*itword].erase(itm);
    }

    sort(context_v.begin(), context_v.end(), comparer);
    string w=*itword;
    lextor_data->set_cooccurrence_context(w, context_v);
  }
}

void 
LexTor::trainlch(istream& is, int left, int right, LexTorData& tlwordmodel, 
                 FSTProcessor& dic, FSTProcessor& bildic, double weigth_exponent) {
  if (lextor_data==NULL) {
    cerr<<"Error in LexTor::trainlch, you must call set_lextor_data before training\n";
    exit(EXIT_FAILURE);
  }

  lextor_data->ensure_stopwords_ok();

  cerr<<"Number of words to take into account on the left side: "<<left<<"\n";
  cerr<<"Number of words to take into account on the right side: "<<right<<"\n";

  set<string> words2workwith=lextor_data->get_words();
  set<string>::iterator itword;

  cerr<<"Words to work with:\n";
  for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
    cerr<<*itword<<"\n";
  }
  cerr<<"\n";

  //For a given lexical choice it stores its translation
  map<string, string> lexchoice_translation;
  map<string, set<string> > lexical_choices_of_word;

  cerr<<"Lexical choices:\n";
  for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
    set<string> lexical_choices=lextor_data->get_lexical_choices(*itword);
    lexical_choices_of_word[*itword]=lexical_choices;
    set<string>::iterator itlch;
    for(itlch=lexical_choices.begin(); itlch!=lexical_choices.end(); itlch++) {
      lexchoice_translation[*itlch]=tlwordmodel.reduce(bildic.biltrans(*itlch,false));
      cerr<<*itlch<<", tr:"<<lexchoice_translation[*itlch]<<"\n";
    }
  }
  cerr<<"\n";


  is.clear();
  is.seekg(ios::beg);

  int nw=0;

  map<string, map<string, COUNT_DATA_TYPE> > context;
  deque<LexTorWord> buffer;

  int word_index=left;
  unsigned buffer_max_size=left+right+1;

  LexTorWord *ltword;
  ltword=LexTorWord::next_word(is,&dic);
  while(ltword!=NULL) {
    //cerr<<"Word read from corpus: "<<ltword->get_word_string()
    //    <<", reduce: "<<lextor_data->reduce(ltword->get_word_string())<<"\n";

    if ((++nw%250000)==0)
      cerr<<nw<<" words processed\n";

    if (!lextor_data->is_stopword(lextor_data->reduce(ltword->get_word_string()))) {	
      if (buffer.size()>=buffer_max_size) {
	buffer.pop_front();
      }
      buffer.push_back(*ltword);

      //The buffer is already full
      if (buffer.size()==buffer_max_size) {

	string reduced_buffer_word=lextor_data->reduce(buffer[word_index].get_word_string());

        for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
	  if (reduced_buffer_word==(*itword)) {
	    //We translate each word in the context
	    //Note: Words in the context can also be ambiguous (with more than one lexical choice)
	    //In that case the count will come from all the possible
	    //translations and divided by the number of possible translations

	    vector<string> translation_buffer[buffer_max_size];
	    string reduced_buffer[buffer_max_size];

	    for (int i=0; i<(int)buffer_max_size; i++) {
	      if (i!=left) 
		reduced_buffer[i]=lextor_data->reduce(buffer[i].get_word_string());	      
	    }

	    //string str_buffer="";
	    //string str_translations="";
	    for (int i=0; i<(int)buffer_max_size; i++) {
	      if ((i!=left)&&(reduced_buffer[i]!=(*itword))) {
		for(int j=0; j<buffer[i].n_lexical_choices(); j++) {
		  string tr=tlwordmodel.reduce(buffer[i].translate(bildic,j));
		  translation_buffer[i].push_back(tr);
		  //str_translations+=tr+"/";
		}
		//str_translations+=" ";
		//str_buffer+=buffer[i].get_word_string()+" ";
	      }
	    }

	    set<string> lexical_choices=lexical_choices_of_word[*itword];
	    set<string>::iterator itlch;
	    /*
	      cerr<<"BUFFER: "<<str_buffer<<"\n";
	      cerr<<"TRANSL: "<<str_translations<<"\n";
	      cerr<<"POLYSEMOUS W: "<<*itword<<"\n";
	      cerr<<"LEXICAL CHOICES: ";
	      for(itlch=lexical_choices.begin(); itlch!=lexical_choices.end(); itlch++) {
	      cerr<<*itlch<<" (TR: "<<lexchoice_translation[*itlch]<<") ";
	      }
	      cerr<<"\n";
	    */

	    map<string, map<string, COUNT_DATA_TYPE> > local_context;
	    COUNT_DATA_TYPE noccurrences=0;

	    //For each lexical choice the counts from the TL are collected
	    for(itlch=lexical_choices.begin(); itlch!=lexical_choices.end(); itlch++) {
	      for (int i=0; i<(int)buffer_max_size; i++) {
		if ((i!=left)&&(reduced_buffer[i]!=(*itword))) {
		  COUNT_DATA_TYPE target_vote=0;

		  for(int j=0; j<(int)translation_buffer[i].size(); j++) {
		    target_vote+=tlwordmodel.vote_from_word(lexchoice_translation[*itlch], 
							    translation_buffer[i][j]);

		    //cerr<<"Target vote from word "<<translation_buffer[i][j]<<" ("
		    //<<lextor_data->reduce(buffer[i].get_word_string())
		    //<<") in tl word model for "<<lexchoice_translation[*itlch]
		    //<<" is "<<new_target_vote<<"\n";
		    
		  }

		  if (target_vote>0) {
		    target_vote=target_vote/((COUNT_DATA_TYPE)translation_buffer[i].size());
		    local_context[*itlch][reduced_buffer[i]]+=target_vote;
		    noccurrences+=target_vote;

		    //context[*itlch][lextor_data->reduce(buffer[i].get_word_string())]+=target_vote;
		    //cerr<<"Count added for "<<*itlch<<" by translations of "
		    //<<lextor_data->reduce(buffer[i].get_word_string())<<": "<<target_vote<<"\n\n";
		  }
		}
	      }
	    }

	    //Now we normalize the counts
	    for(itlch=lexical_choices.begin(); itlch!=lexical_choices.end(); itlch++) {
	      int distance=(-1)*left;
	      for (int i=0; i<(int)buffer_max_size; i++) { 
		if ((i!=left)&&(reduced_buffer[i]!=(*itword))) {
		  if (local_context[*itlch][reduced_buffer[i]]>0) {
		    double count=local_context[*itlch][reduced_buffer[i]]/noccurrences;
		    context[*itlch][reduced_buffer[i]]+=count/pow(fabs((double)distance),weigth_exponent);
		  }
		}
		distance++;
	      }
	    }

	    break;
	  }
	}
      }
    } 

    delete ltword;
    ltword=LexTorWord::next_word(is,&dic);
  }
  
  cerr<<"Corpus has "<<nw<<" words\n";

  //All co-occurrences have been collected. We need to filter them
  //so as to take into account only the n most frequent
  for(itword=words2workwith.begin(); itword!=words2workwith.end(); itword++) {
    set<string> lexical_choices=lexical_choices_of_word[*itword];
    set<string>::iterator itlch;
    for(itlch=lexical_choices.begin(); itlch!=lexical_choices.end(); itlch++) {
      PairStringCountComparer comparer;
      vector<pair<string, COUNT_DATA_TYPE> > context_v;
      map<string, COUNT_DATA_TYPE>::iterator itm;

      while(context[*itlch].size()>0) {
	itm=context[*itlch].begin();
	//cerr<<itm->first<<" "<<itm->second<<"\n";
	context_v.push_back(*itm);
	context[*itlch].erase(itm);
      }
    

      sort(context_v.begin(), context_v.end(), comparer);
      string lch=*itlch;
      lextor_data->set_cooccurrence_context(lch, context_v);
    }
  } 

  cerr<<"Training done\n"; 
}

void 
LexTor::lexical_selector(istream& is, FSTProcessor &fstp, int left, int right, double weigth_exponent, LexTorEval* lteval) {
  if (lextor_data==NULL) {
    cerr<<"Error in LexTor::lexical_selector, you must call set_lextor_data before\n";
    exit(EXIT_FAILURE);
  }

  deque<LexTorWord> buffer;
  deque<LexTorWord> window;

  LexTorWord  nullword("NULLWORD", &fstp);

  for(int i=0; i<(left+right+1); i++)
    window.push_back(nullword);

  int retain=0;

  LexTorWord* ltword=NULL;
  ltword=LexTorWord::next_word(is, &fstp);

  /*
    cerr<<"WINDOW: ";
    for(int i=0; i<(int)window.size(); i++)
    cerr<<window[i].get_word_string()<<" ";
    cerr<<"\n";
    cerr<<"BUFFER: ";
    for(int i=0; i<(int)buffer.size(); i++)
    cerr<<buffer[i].get_word_string()<<" ";
    cerr<<"\n\n";
  */

  while(ltword) {
    //cerr<<"Word read: "<<ltword->get_word_string()<<", reduced: "<<lextor_data->reduce(ltword->get_word_string())<<" ";
    //cerr<<"# lexical choices: "<<ltword->n_lexical_choices()<<"\n";

    if (!lextor_data->is_stopword(lextor_data->reduce(ltword->get_word_string()))) { 
      //cerr<<"  Not stopword, retain="<<retain<<"\n";
      if (window.size()>=(unsigned)(left+1+right)) 
	window.pop_front();
      
      window.push_back(*ltword);

      if (ltword->n_lexical_choices()>1) {
	retain++;
	if (retain>1)
	  buffer.push_back(*ltword);
      } else {
	if (retain>0) 
	  buffer.push_back(*ltword);
	else {
	  //cout<<ltword->get_lexical_choice(-1,false)<<"\n";
	  cout<<ltword->get_lexical_choice(-1,true);
	  if (lteval) 
	    lteval->evalword(*ltword, -1, lextor_data);
	  //cerr<<"OUT: "<<ltword->get_lexical_choice()<<"\n";
	}
      }

      if (window[left].n_lexical_choices()>1) {
	int winner=estimate_winner_lch(window, left, weigth_exponent);

	//cerr<<"WINNER: "<<winner<<"\n";
	//cerr<<"OUT:"<<window[left].get_lexical_choice(winner)<<"\n";

	//cout<<window[left].get_lexical_choice(winner,false);
	cout<<window[left].get_lexical_choice(winner,true);
	if (lteval) 
	  lteval->evalword(window[left], winner, lextor_data);
	
	//For debug
	/*
	  cout<<" | ";
	  for(int j=0; j<window[left].n_lexical_choices(); j++) {
	  if (j>0)
	  cout<<"|";
	  cout<<window[left].get_lexical_choice(j,false);
	  }
	  cout<<"\n";
	*/
	//Now those words that were retained must be released
	if(retain>0) {
	  while ((buffer.size()>0)&&(buffer[0].n_lexical_choices()==1)) {
	    //cout<<buffer[0].get_lexical_choice(-1,false)<<"\n";
	    cout<<buffer[0].get_lexical_choice(-1,true);
	    if (lteval) 
	      lteval->evalword(buffer[0], -1, lextor_data);
	    //cerr<<"OUT: "<<buffer[0].get_lexical_choice()<<"\n";
	    buffer.pop_front();
	  }
	  if ((buffer.size()>0)&&(buffer[0].n_lexical_choices()>1))
	    buffer.pop_front(); 

	  retain--;
	}
      } 
    } else { //It's a stopword
      //cerr<<"  Is stopword, retain="<<retain<<"\n";
      if (retain>0) 
	buffer.push_back(*ltword);
      else {
	//cout<<ltword->get_lexical_choice(-1,false)<<"\n";
	cout<<ltword->get_lexical_choice(-1,true);
	if (lteval) 
	  lteval->evalword(*ltword, -1, lextor_data);
	//cerr<<"OUT: "<<ltword->get_lexical_choice()<<"\n";
      }
    }
    /*
      cerr<<"WINDOW: ";
      for(int i=0; i<(int)window.size(); i++)
      cerr<<window[i].get_word_string()<<" ";
      cerr<<"\n";
      cerr<<"BUFFER: ";
      for(int i=0; i<(int)buffer.size(); i++)
      cerr<<buffer[i].get_word_string()<<" ";
      cerr<<"\n\n";
    */
    delete ltword;
    ltword=LexTorWord::next_word(is, &fstp);
  }

  if (retain>0) {
    for(unsigned i=left+1; i<window.size(); i++) {
      if (window[i].n_lexical_choices()>1) {
	int winner=estimate_winner_lch(window, i, weigth_exponent);

	//cerr<<"WINNER: "<<winner<<"\n";
	//cerr<<"OUT: "<<window[i].get_lexical_choice(winner)<<"\n";

	//cout<<window[i].get_lexical_choice(winner,false);
	cout<<window[i].get_lexical_choice(winner,true);
	if (lteval)
	  lteval->evalword(window[i], winner, lextor_data);

	//For debug
	/*
	  cout<<" | ";
	  for(int j=0; j<window[i].n_lexical_choices(); j++) {
	  if (j>0)
	  cout<<"|";
	  cout<<window[i].get_lexical_choice(j,false);
	  }
	  cout<<"\n";
	*/

	//Now those words that were retained must be released
	if(retain>0) {
	  while ((buffer.size()>0)&&(buffer[0].n_lexical_choices()==1)) {
	    //cout<<buffer[0].get_lexical_choice(-1,false)<<"\n";
	    cout<<buffer[0].get_lexical_choice(-1,true);
	    if (lteval) 
	      lteval->evalword(buffer[0], -1, lextor_data);
	    //cerr<<"OUT: "<<buffer[0].get_lexical_choice()<<"\n";
	    buffer.pop_front();
	  }
	  if ((buffer.size()>0)&&(buffer[0].n_lexical_choices()>1))
	    buffer.pop_front(); 

	  retain--;
	}
	/*
	  cerr<<"WINDOW: ";
	  for(int i=0; i<(int)window.size(); i++)
	  cerr<<window[i].get_word_string()<<" ";
	  cerr<<"\n";
	  cerr<<"BUFFER: ";
	  for(int i=0; i<(int)buffer.size(); i++)
	  cerr<<buffer[i].get_word_string()<<" ";
	  cerr<<"\n\n";
	*/
      }
    }
  }

  //cerr<<"retain: "<<retain<<"\n";
}

int 
LexTor::estimate_winner_lch(deque<LexTorWord>& window, int word_index, double weigth_exponent) {
  COUNT_DATA_TYPE lexchoices_count[window[word_index].n_lexical_choices()];

  if (debug) {
    cerr<<"WINDOW\n------------------\n";
    for(unsigned i=0; i<window.size(); i++)
      cerr<<lextor_data->reduce(window[i].get_word_string())<<" ";
    cerr<<"\n";
    cerr<<"WORD INDEX: "<<word_index<<" "
	<<lextor_data->reduce(window[word_index].get_word_string())<<"\n";
  }

  for(int i=0; i<window[word_index].n_lexical_choices(); i++) {
    lexchoices_count[i]=0;
    string reduced_lexchoice=lextor_data->reduce_lexical_choice(window[word_index].get_lexical_choice(i,false));
    if (debug) {
      cerr<<"lexical choice: "<<window[word_index].get_lexical_choice(i)<<" reduced: "<<reduced_lexchoice<<"\n";
    }

    int distance=(-1)*(word_index);
    for(int j=0; j<(int)window.size(); j++) { 
      //For all words in the context window
      if(j!=word_index) {
	COUNT_DATA_TYPE vote=0;

	string reduced_word=lextor_data->reduce(window[j].get_word_string());

	if (lextor_data->get_wordcount(reduced_word)>0) {
	  vote=lextor_data->vote_from_word(reduced_lexchoice, reduced_word)/
	    lextor_data->get_wordcount(reduced_word);
	  lexchoices_count[i]+=vote/pow(fabs((double)distance),weigth_exponent);
	}
	if (debug) {
	  cerr<<"Count for "<<reduced_lexchoice<<" from "<<reduced_word<<" is "<<vote<<"\n";
	  cerr<<"Vote: "<<lextor_data->vote_from_word(reduced_lexchoice, reduced_word)<<" word count: "
	      <<lextor_data->get_wordcount(reduced_word)<<"\n";
	}
      }
      distance++;
    }
  }

  //Now the winner is calculated
  int winner=-1; //This will make the default one to be used if unchanged
  COUNT_DATA_TYPE winner_vote=-100000000;
  for(int i=0; i<window[word_index].n_lexical_choices(); i++) {
    if ((lexchoices_count[i]>0) && (lexchoices_count[i]>winner_vote)) {
      winner_vote=lexchoices_count[i];
      winner=i;
    } else if ((lexchoices_count[i]>0) && (lexchoices_count[i]==winner_vote)) {
      //Take the most probable one, the one with the highest sum
      COUNT_DATA_TYPE sum_i=lextor_data->get_lexchoice_sum(lextor_data->reduce(window[word_index].get_lexical_choice(i)));
      COUNT_DATA_TYPE sum_win=lextor_data->get_lexchoice_sum(lextor_data->reduce(window[word_index].get_lexical_choice(winner)));
      if (sum_i>sum_win)
	winner=i;
    }
  }

  if (debug) {
    cerr<<"WINNER: "<<winner<<" "<<window[word_index].get_lexical_choice(winner)<<"\n";
  }
  return winner;
}

void
LexTor::print_debug(const string& s) {
  if (debug)
    cerr<<s;
}

void 
LexTor::print_debug(const double& d){
  if (debug)
    cerr<<d;
}
void 
LexTor::print_debug(const int& i) {
  if (debug)
    cerr<<i; 
}
